# Copyright 2021 The CLU Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for clu.metrics."""

import functools
import operator
from unittest import mock

from absl.testing import parameterized
from clu import metrics
import flax
import jax
import jax.numpy as jnp
import tensorflow as tf


@flax.struct.dataclass
class CollectingMetricAccuracy(
    metrics.CollectingMetric.from_outputs(("logits", "labels"))):

  def compute(self):
    logits = self.values["logits"]
    labels = self.values["labels"]
    assert logits.ndim == 2, logits.shape
    assert labels.ndim == 1, labels.shape
    return (logits.argmax(axis=-1) == labels).mean()


@flax.struct.dataclass
class Collection(metrics.Collection):
  train_accuracy: metrics.Accuracy
  collecting_metric_accuracy: CollectingMetricAccuracy
  learning_rate: metrics.LastValue.from_output("learning_rate")
  collecting_metric: metrics.CollectingMetric.from_outputs(("logits", "labels"))


class MetricsTest(tf.test.TestCase, parameterized.TestCase):

  def setUp(self):
    super().setUp()
    # Two batches of model output.
    self.model_outputs = (
        dict(
            logits=jnp.array([[1., 0.], [0., 1.]]),
            labels=jnp.array([0, 0]),
            learning_rate=0.02,
            loss=jnp.array(4.2),
        ),
        dict(
            logits=jnp.array([[1., 2.], [3., 4.]]),
            labels=jnp.array([1, 1]),
            learning_rate=0.01,
            loss=jnp.array(1.7),
        ),
    )
    masks = (
        jnp.array([False, True]),
        jnp.array([True, False]),
    )
    self.model_outputs_masked = tuple(
        dict(mask=mask, **model_output)
        for mask, model_output in zip(masks, self.model_outputs))
    def concat_outputs(name):
      return jnp.concatenate(
          [model_output[name] for model_output in self.model_outputs_masked])

    self.results = {
        "train_accuracy": 0.75,
        "learning_rate": 0.01,
        "collecting_metric_accuracy": 0.75,
        "collecting_metric": {
            "labels": concat_outputs("labels"),
            "logits": concat_outputs("logits"),
        },
    }
    self.results_masked = {
        "train_accuracy": 0.5,
        "learning_rate": 0.01,
        # Note: Our CollectingMetric does NOT collect the mask (the masked
        # computation couldn't be jitted anyways).
        "collecting_metric_accuracy": 0.75,
        "collecting_metric": {
            "labels": concat_outputs("labels"),
            "logits": concat_outputs("logits"),
        },
    }
    # Stack all values. Can for example be used in a pmap().
    self.model_outputs_stacked = jax.tree_multimap(
        lambda *args: jnp.stack(args), *self.model_outputs)
    self.model_outputs_masked_stacked = jax.tree_multimap(
        lambda *args: jnp.stack(args), *self.model_outputs_masked)

  def make_compute_metric(self, metric_class, reduce):
    """Returns a jitted function to compute metrics.

    Args:
      metric_class: Metric class to instantiate.
      reduce: If set to `True`

    Returns:
      A function that takes `model_outputs` (list of dictionaries of values) as
      an input and returns the value from `metric.compute()`.
    """

    @jax.jit
    def compute_metric(model_outputs):
      if reduce:
        metric_list = [
            metric_class.from_model_output(**model_output)
            for model_output in self.model_outputs
        ]
        metric_stacked = jax.tree_multimap(lambda *args: jnp.stack(args),
                                           *metric_list)
        metric = metric_stacked.reduce()
      else:
        metric = None
        for model_output in model_outputs:
          update = metric_class.from_model_output(**model_output)
          metric = update if metric is None else metric.merge(update)
      return metric.compute()

    return compute_metric

  def test_metric_reduce(self):
    metric1 = metrics.LastValue.from_model_output(jnp.array([1, 2]))
    metric2 = metrics.LastValue.from_model_output(jnp.array([3, 4]))
    metric12 = jax.tree_multimap(lambda *args: jnp.stack(args), metric1,
                                 metric2)
    self.assertAllEqual(metric12.reduce().compute(), metric2.compute())

  @parameterized.named_parameters(
      ("", False),)
  def test_average_fun(self, reduce):

    def accuracy(*, logits, labels, **_):
      return (logits.argmax(axis=-1) == labels).astype(jnp.float32)

    self.assertAllClose(
        self.make_compute_metric(metrics.Average.from_fun(accuracy),
                                 reduce)(self.model_outputs), 0.75)

  @parameterized.named_parameters(
      ("Average", metrics.Average),
      ("Std", metrics.Std),
      ("LastValue", metrics.LastValue),
  )
  def test_merge_asserts_shape(self, metric_cls):
    metric1 = metric_cls.from_model_output(jnp.arange(3.))
    metric2 = jax.tree_multimap(lambda *args: jnp.stack(args), metric1, metric1)
    with self.assertRaisesRegex(ValueError, r"^Expected same shape"):
      metric1.merge(metric2)

  @parameterized.named_parameters(
      ("", False),
      ("_reduce", True),
  )
  def test_accuracy(self, reduce):
    self.assertAllClose(
        self.make_compute_metric(metrics.Accuracy, reduce)(self.model_outputs),
        0.75)

  def test_last_value_asserts_shape(self):
    metric1 = metrics.LastValue.from_model_output(jnp.arange(3.))
    metric2 = jax.tree_multimap(lambda *args: jnp.stack(args), metric1, metric1)
    with self.assertRaisesRegex(ValueError, r"^Expected same shape"):
      metric1.merge(metric2)

  @parameterized.named_parameters(
      ("", False),
      ("_reduce", True),
  )
  def test_loss_average(self, reduce):
    self.assertAllClose(
        self.make_compute_metric(metrics.Average.from_output("loss"),
                                 reduce)(self.model_outputs),
        self.model_outputs_stacked["loss"].mean())

  @parameterized.named_parameters(
      ("", False),
      ("_reduce", True),
  )
  def test_loss_std(self, reduce):
    self.assertAllClose(
        self.make_compute_metric(metrics.Std.from_output("loss"),
                                 reduce)(self.model_outputs),
        self.model_outputs_stacked["loss"].std(),
        atol=1e-4)

  @parameterized.named_parameters(
      ("", False),
      ("_masked", True),
  )
  def test_collection_single(self, masked):

    def compute_collection(model_outputs):
      collection = None
      for model_output in model_outputs:
        update = Collection.single_from_model_output(**model_output)
        collection = (
            update if collection is None else collection.merge(update))
      return collection.compute()

    self.assertAllClose(
        jax.jit(compute_collection)(
            self.model_outputs_masked if masked else self.model_outputs),
        self.results_masked if masked else self.results)

  @parameterized.named_parameters(
      ("", False),
      ("_masked", True),
  )
  @mock.patch("jax.lax.all_gather")
  def test_collection_gather(self, masked, all_gather_mock):

    collections = [
        Collection.single_from_model_output(**model_output)
        for model_output in (
            self.model_outputs_masked if masked else self.model_outputs)
    ]
    all_gather_mock.return_value = jax.tree_multimap(
        lambda *args: jnp.stack(args), *collections)

    def compute_collection(model_outputs):
      collection = Collection.gather_from_model_output(**model_outputs[0])
      return collection.compute()

    self.assertAllClose(
        jax.jit(compute_collection)(
            self.model_outputs_masked if masked else self.model_outputs),
        self.results_masked if masked else self.results)

  @parameterized.named_parameters(
      ("", False),
      ("_masked", True),
  )
  def test_collection_gather_pmap(self, masked):

    @functools.partial(jax.pmap, axis_name="batch")
    def compute_collection(model_outputs):
      return Collection.gather_from_model_output(**model_outputs).compute()

    if jax.local_device_count() > 1:
      self.assertAllClose(
          flax.jax_utils.unreplicate(
              compute_collection(self.model_outputs_masked_stacked
                                 if masked else self.model_outputs_stacked)),
          self.results_masked if masked else self.results)

  def test_collection_asserts_replication(self):
    collections = [
        Collection.single_from_model_output(**model_output)
        for model_output in self.model_outputs
    ]
    collection = jax.tree_multimap(lambda *args: jnp.stack(args), *collections)
    with self.assertRaisesRegex(ValueError,
                                r"^Collection is still replicated"):
      collection.compute()

  def test_collecting_metric(self):
    metric_class = metrics.CollectingMetric.from_outputs(("logits", "loss"))
    logits = jnp.concatenate(
        [model_output["logits"] for model_output in self.model_outputs])
    loss = jnp.array(
        [model_output["loss"] for model_output in self.model_outputs])
    result = self.make_compute_metric(
        metric_class, reduce=False)(
            self.model_outputs)
    self.assertAllClose(result, {
        "logits": logits,
        "loss": loss,
    })

  def test_collecting_metric_reduce(self):
    metric_class = metrics.CollectingMetric.from_outputs(("logits", "loss"))
    logits = functools.reduce(
        operator.add,
        ((model_output["logits"],) for model_output in self.model_outputs))
    loss = functools.reduce(
        operator.add,
        ((model_output["loss"],) for model_output in self.model_outputs))
    result = self.make_compute_metric(
        metric_class, reduce=True)(
            self.model_outputs)
    # Expected : single tuple with concatenated elements.
    self.assertAllClose(result, {
        "logits": jnp.concatenate(logits),
        "loss": jnp.stack(loss),
    })


if __name__ == "__main__":
  tf.test.main()



# Copyright 2021 The CLU Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MetricWriter for Pytorch summary files.

Use this writer for the Pytorch-based code.

"""

from typing import Any, Mapping, Optional


from clu.metric_writers import interface
import numpy as np
from torch.utils import tensorboard


Scalar = interface.Scalar


class TorchTensorboardWriter(interface.MetricWriter):
  """MetricWriter that writes Pytorch summary files."""

  def __init__(self, logdir: str):
    super().__init__()
    self._writer = tensorboard.SummaryWriter(log_dir=logdir)


  def write_scalars(self, step: int, scalars: Mapping[str, Scalar]):
    for key, value in scalars.items():
      self._writer.add_scalar(key, value, global_step=step)

  def write_images(self, step: int, images: Mapping[str, np.ndarray]):
    for key, value in images.items():
      self._writer.add_image(key, value, global_step=step, dataformats="HWC")

  def write_texts(self, step: int, texts: Mapping[str, str]):
    for key, value in texts.items():
      self._writer.text(key, value, global_step=step)

  def write_histograms(self,
                       step: int,
                       arrays: Mapping[str, np.ndarray],
                       num_buckets: Optional[Mapping[str, int]] = None):
    for tag, values in arrays.items():
      bins = None if num_buckets is None else num_buckets.get(tag)
      self._writer.add_histogram(
          tag, values, global_step=step, bins="auto", max_bins=bins)

  def write_hparams(self, hparams: Mapping[str, Any]):
    self._writer.add_hparams(hparams, {})

  def flush(self):
    self._writer.flush()

  def close(self):
    self._writer.close()



# Copyright 2021 The CLU Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""setup.py for Common Loop Utils.

Install for development:

  pip intall -e . .[tests]
"""

from setuptools import find_packages
from setuptools import setup

tests_require = [
    "dm-sonnet",
    "pytest",
    "torch>=1.2.0",
]
pytorch_require = [
    "torch>=1.2.0",
]

setup(
    name="clu",
    version="0.0.2",
    description=("Set of libraries for ML training loops in JAX."),
    author="Common Loop Utils Authors",
    author_email="no-reply@google.com",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="http://github.com/google/CommonLoopUtils",
    license="Apache 2.0",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "absl-py",
        "dataclasses;python_version<'3.7'",
        "flax",
        "jax>=0.1.76",
        "jaxlib",
        "ml_collections",
        "numpy>=1.16.4",
        "tensorflow>=2.3.0",
        "tensorflow_datasets",
    ],
    tests_require=tests_require,
    extras_require=dict(test=tests_require, pytorch=pytorch_require),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="JAX machine learning",
)



# Copyright 2021 The CLU Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the LoggingWriter."""

from clu.metric_writers import logging_writer
import numpy as np
import tensorflow as tf


class LoggingWriterTest(tf.test.TestCase):

  def setUp(self):
    super().setUp()
    self.writer = logging_writer.LoggingWriter()

  def test_write_scalars(self):
    with self.assertLogs(level="INFO") as logs:
      self.writer.write_scalars(0, {"a": 3, "b": 0.15})
      self.writer.write_scalars(2, {"a": 5, "b": 0.007})
    self.assertEqual(
        logs.output,
        ["INFO:absl:[0] a=3, b=0.150000", "INFO:absl:[2] a=5, b=0.007000"])

  def test_write_images(self):
    images = np.zeros((2, 28, 28, 3))
    with self.assertLogs(level="INFO") as logs:
      self.writer.write_images(4, {"input_images": images})
    self.assertEqual(
        logs.output,
        ["INFO:absl:[4] Got images: {'input_images': (2, 28, 28, 3)}."])

  def test_write_texts(self):
    with self.assertLogs(level="INFO") as logs:
      self.writer.write_texts(4, {"samples": "bla"})
    self.assertEqual(
        logs.output,
        ["INFO:absl:[4] Got texts: {'samples': 'bla'}."])

  def test_write_histogram(self):
    with self.assertLogs(level="INFO") as logs:
      self.writer.write_histograms(
          step=4,
          arrays={
              "a": np.asarray([-0.1, 0.1, 0.3]),
              "b": np.arange(31),
              "c": np.asarray([0.1, 0.1, 0.1, 0.1, 0.1]),
          },
          num_buckets={
              "a": 2,
              "c": 1
          })
    # Note: There are 31 distinct values [0, 1, ..., 30], and 30 buckets by
    # default. Last bucket gets 2 values.
    expected_histo_b = ", ".join([f"[{i}, {i + 1}): 1" for i in range(29)] +
                                 ["[29, 30]: 2"])
    self.assertEqual(logs.output, [
        "INFO:absl:[4] Histogram for 'a' = {[-0.1, 0.1): 1, [0.1, 0.3]: 2}",
        f"INFO:absl:[4] Histogram for 'b' = {{{expected_histo_b}}}",
        "WARNING:absl:num_buckets was automatically changed from 1 to 2",
        "INFO:absl:[4] Histogram for 'c' = {[-0.4, 0.6]: 5}",
    ])

  def test_write_hparams(self):
    with self.assertLogs(level="INFO") as logs:
      self.writer.write_hparams({"learning_rate": 0.1, "batch_size": 128})
    self.assertEqual(logs.output, [
        "INFO:absl:Hyperparameters: {'learning_rate': 0.1, 'batch_size': 128}"
    ])


if __name__ == "__main__":
  tf.test.main()



# Copyright 2021 The CLU Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple checkpointing library for TF2/Flax.

The class `Checkpoint` is a simple wrapper around `tf.train.Checkpoint` that
also stores a `flax.struct.dataclass` instance in the same directory.

Synopsis:

  from clu import checkpoint
  import flax

  @flax.struct.dataclass
  class TrainState:
    optimizer: flax.optim.Optimizer
    step: int

  ds = load_tf_dataset()
  ds_iter = iter(ds)
  ckpt = checkpoint.MultihostCheckpoint(base_directory, dict(ds_iter=ds_iter))
  optimizer = create_flax_optimizer()
  state = TrainState(optimizer=optimizer, step=0)
  state = ckpt.restore_or_initialize(state)  # Also restores `ds_iter`.
  initial_step = int(state.step) + 1
  # Need to replicate all data when training with multiple accelerators.
  state = flax.jax_utils.replicate(state)

  for step in range(initial_step, steps + 1):
    state = update_step(state, next(ds_iter))
    ckpt.save(flax.jax_utils.unreplicate(state))

Loading the model e.g. in a Colab:

  from clu import checkpoint
  import flax
  from . import mnist_lib

  state_dict = checkpoint.load_state_dict(base_directory)
  params = state_dict['optimizer']['target']['params']
  module = mnist_lib.MyArchitecture.partial(num_classes=10)
  model = flax.nn.Model(module, params)
"""

import collections
import re
from typing import Any, Dict, Optional, TypeVar

from absl import logging

from clu.internal import utils
import flax
import jax
import tensorflow as tf




T = TypeVar("T")


def load_state_dict(base_directory) -> Dict[str, Any]:
  """Restores `state` as dictionary from the latest checkpoint.

  Synopsis:

    data = checkpoint.load_state_dict(base_directory)
    params = data['optimizer']['target']['params']
    module = mnist_lib.MyArchitecture.partial(num_classes=10)
    model = flax.nn.Model(module, params)

  Args:
    base_directory: Directory from which the checkpoints should be restored. See
      `Checkpoint.__init__()`.

  Returns:
    The deserialized Flax data, as a dictionary.

  Raises:
    FileNotFoundError: If there is no checkpoint to restore.
  """
  ckpt = Checkpoint(base_directory)
  if not ckpt.latest_checkpoint:
    raise FileNotFoundError(f"No checkpoint found in {base_directory}")
  with utils.log_activity("load_state_dict"):
    with tf.io.gfile.GFile(ckpt.latest_checkpoint_flax, "rb") as f:
      return flax.serialization.msgpack_restore(f.read())


class CheckpointInfo(
    collections.namedtuple("CheckpointInfo", ("prefix", "number"))):
  """Helper class to parse a TensorFlow checkpoint path."""

  CHECKPOINT_REGEX = r"^(?P<prefix>.*)-(?P<number>\d+)"

  @classmethod
  def initialize(cls, base_directory, checkpoint_name: str) -> "CheckpointInfo":
    """Creates a first CheckpointInfo (number=1)."""
    return cls(f"{base_directory}/{checkpoint_name}", 1)

  @classmethod
  def from_path(cls, checkpoint: str) -> "CheckpointInfo":
    """Parses a checkpoint.

    Args:
      checkpoint: A checkpoint prefix, as can be found in the
        `.latest_checkpoint` property of a `tf.train.CheckpointManager`.

    Returns:
      An instance of `CheckpointInfo` that represents `checkpoint`.
    """
    m = re.match(cls.CHECKPOINT_REGEX, checkpoint)
    if m is None:
      RuntimeError(f"Invalid checkpoint format: {checkpoint}")
    d = m.groupdict()  # pytype: disable=attribute-error
    return cls(d["prefix"], int(d["number"]))

  def increment(self) -> "CheckpointInfo":
    """Returns a new CheckpointInfo with `number` increased by one."""
    return CheckpointInfo(self.prefix, self.number + 1)

  def __str__(self):
    """Does the opposite of `.from_path()`."""
    return f"{self.prefix}-{self.number}"


class Checkpoint:
  """A utility class for storing and loading TF2/Flax checkpoints.

  Both the state of a `tf.data.Dataset` iterator and a `flax.struct.dataclass`
  are stored on disk in the following files:

  - {directory}/checkpoint
  - {directory}/ckpt-{number}.index
  - {directory}/ckpt-{number}.data@*
  - {directory}/ckpt-{number}.flax

  Where {number} starts at 1 is then incremented by 1 for every new checkpoint.
  The last file is the `flax.struct.dataclass`, serialized in Messagepack
  format. The other files are explained in more detail in the Tensorflow
  documentation:

  https://www.tensorflow.org/api_docs/python/tf/train/Checkpoint
  """

  def __init__(self,
               base_directory: str,
               tf_state: Optional[Dict[str, Any]] = None,
               *,
               max_to_keep: int = 5,
               checkpoint_name: str = "ckpt"):
    """Initializes a Checkpoint with a dictionary of TensorFlow Trackables.

    Args:
      base_directory: Directory under which the checkpoints will be stored. Use
        a different base_directory in every task.
      tf_state: A dictionary of TensorFlow `Trackable` to be serialized, for
        example a dataset iterator.
      max_to_keep: Number of checkpoints to keep in the directory. If there are
        more checkpoints than specified by this number, then the oldest
        checkpoints are removed.
      checkpoint_name: Prefix of the checkpoint files (before `-{number}`).
    """
    if tf_state is None:
      tf_state = dict()
    self.base_directory = base_directory
    self.max_to_keep = max_to_keep
    self.checkpoint_name = checkpoint_name
    self.tf_checkpoint = tf.train.Checkpoint(**tf_state)
    self.tf_checkpoint_manager = tf.train.CheckpointManager(
        self.tf_checkpoint,
        base_directory,
        max_to_keep=max_to_keep,
        checkpoint_name=checkpoint_name)

  def get_latest_checkpoint_to_restore_from(self):
    """Returns the latest checkpoint to restore from.

    In the current implementation, this method simply returns the attribute
    `latest_checkpoint`.

    Subclasses can override this method to provide an alternative checkpoint to
    restore from, for example for synchronization across multiple checkpoint
    directories.
    """
    return self.latest_checkpoint

  @property
  def latest_checkpoint(self) -> Optional[str]:
    """Latest checkpoint, see `tf.train.CheckpointManager.latest_checkpoint`.

    Returns:
      A string to the latest checkpoint. Note that this string is path-like but
      it does not really describe a file, but rather a set of files that are
      constructed from this string, by appending different file extensions. The
      returned value is `None` if there is no previously stored checkpoint in
      `base_directory` specified to `__init__()`.
    """
    return self.tf_checkpoint_manager.latest_checkpoint

  @property
  def latest_checkpoint_flax(self) -> Optional[str]:
    """Path of the latest serialized `state`.

    Returns:
      Path of the file containing the serialized Flax state. The returned value
      is `None` if there is no previously stored checkpoint in `base_directory`
      specified to `__init__()`.
    """
    if self.latest_checkpoint is None:
      return None
    return self._flax_path(self.latest_checkpoint)

  def _flax_path(self, checkpoint: str) -> str:
    return "{}.flax".format(checkpoint)

  def _next_checkpoint(self, checkpoint: Optional[str]) -> str:
    if checkpoint is None:
      return str(
          CheckpointInfo.initialize(self.base_directory, self.checkpoint_name))
    return str(CheckpointInfo.from_path(checkpoint).increment())

  def _checkpoint_number(self, checkpoint: Optional[str]) -> Optional[int]:
    if checkpoint is not None:
      return CheckpointInfo.from_path(checkpoint).number

  @utils.logged_with("Checkpoint.save()")
  def save(self, state) -> str:
    """Saves a new checkpoints in the directory.

    Args:
      state: Flax checkpoint to be stored.

    Returns:
      The checkpoint identifier ({base_directory}/ckpt-{number}).

    Raises:
      RuntimeError: If tf_checkpoint.save_counter does not match
          tf_checkpoint_manager.latest_checkpoint.
    """
    latest_checkpoint_num = self._checkpoint_number(self.latest_checkpoint) or 0
    if latest_checkpoint_num != self.tf_checkpoint.save_counter.numpy():
      raise RuntimeError(
          f"Expected save_counter={self.tf_checkpoint.save_counter.numpy()} "
          f"to match latest_checkpoint={self.latest_checkpoint}. Make sure "
          f"the checkpoint is initialized via `.restore_or_initialize()` "
          f"before it's stored and that no other process writes to the same "
          f"checkpoint directory.")
    next_checkpoint = self._next_checkpoint(self.latest_checkpoint)
    flax_path = self._flax_path(next_checkpoint)
    if not tf.io.gfile.exists(self.base_directory):
      tf.io.gfile.makedirs(self.base_directory)
    with tf.io.gfile.GFile(flax_path, "wb") as f:
      f.write(flax.serialization.to_bytes(state))
    checkpoints = set(self.tf_checkpoint_manager.checkpoints)
    # Write Tensorflow data last. This way Tensorflow checkpoint generation
    # logic will make sure to only commit checkpoints if they complete
    # successfully. A previously written `flax_path` would then simply be
    # overwritten next time.
    self.tf_checkpoint_manager.save()
    for checkpoint in checkpoints.difference(
        self.tf_checkpoint_manager.checkpoints):
      tf.io.gfile.remove(self._flax_path(checkpoint))
    if next_checkpoint != self.latest_checkpoint:
      raise AssertionError(  # pylint: disable=g-doc-exception
          "Expected next_checkpoint to match latest_checkpoint: "
          f"{next_checkpoint} != {self.latest_checkpoint}")
    return self.latest_checkpoint

  @utils.logged_with("Checkpoint.restore_or_initialize()")
  def restore_or_initialize(self, state: T) -> T:
    """Restores from the latest checkpoint, or creates a first checkpoint.

    Args:
      state : A flax checkpoint to be stored or to serve as a template. If the
        checkoint is restored (and not initialized), then the fields of `state`
        must match the data previously stored.

    Returns:
      The restored `state` object. Note that all TensorFlow `Trackable`s in
      `tf_state` (see `__init__()`) are also updated.
    """
    latest_checkpoint = self.get_latest_checkpoint_to_restore_from()
    if not latest_checkpoint:
      logging.info("No previous checkpoint found.")
      self.save(state)
      return state
    logging.info("Found previous checkpoint: %s", latest_checkpoint)
    self.tf_checkpoint.restore(latest_checkpoint)
    flax_path = self._flax_path(latest_checkpoint)
    with tf.io.gfile.GFile(flax_path, "rb") as f:
      state = flax.serialization.from_bytes(state, f.read())
    return state

  def restore(self, state: T) -> T:
    """Restores from the latest checkpoint.

    Similar to `restore_or_initialize()`, but raises a `FileNotFoundError` if
    there is no checkpoint.

    Args:
      state : A flax checkpoint to be stored or to serve as a template. If the
        checkoint is restored (and not initialized), then the fields of `state`
        must match the data previously stored.

    Returns:
      The restored `state` object. Note that all TensorFlow `Trackable`s in
      `tf_state` (see `__init__()`) are also updated.

    Raises:
      FileNotFoundError: If there is no checkpoint to restore.
    """
    latest_checkpoint = self.get_latest_checkpoint_to_restore_from()
    if not latest_checkpoint:
      raise FileNotFoundError(f"No checkpoint found at {self.base_directory}")
    return self.restore_or_initialize(state)


class MultihostCheckpoint(Checkpoint):
  """An subclass of `Checkpoint` that synchronizes between multiple JAX hosts.

  If the training split across multiple hosts, then the following race condition
  can occur : If a host is pre-empted while writing a checkpoint, then the other
  hosts will only be restarted with a small delay, and at that point they
  probably already have finished writing their checkpoint. Upon restart, the
  host that was interrupted while writing the checkpoint will load the latest
  fully written checkpoint, which will be out of sync with the other hosts that
  successfully wrote one more checkpoint.

  This class also allows to specify a `multihost_base_directory` that is
  identical for all hosts and will be used to drive a host-specific directory.
  """

  def __init__(self,
               multihost_base_directory: str,
               tf_state: Optional[Dict[str, Any]] = None,
               *,
               host_id: Optional[int] = None,
               max_to_keep: int = 5,
               checkpoint_name: str = "ckpt"):
    """Initializes a MultihostCheckpoint with a dict of TensorFlow Trackables.

    Args:
      multihost_base_directory: Directory that will be used to construct a
        host-specific `base_directory` under which the checkpoints will be
        stored. Usually a directory *within* the work unit's workdirectory
        (e.g. `f"{workdir}/checkpoints`). One directory per host will be created
        at the same level as this base directory labeled
        `f"{multihost_base_directory}-{host_id}"`.
      tf_state: A dictionary of TensorFlow `Trackable` to be serialized, for
        example a dataset iterator.
      host_id: Host ID used to construct the `base_directory`. Taken from
        `jax.host_id()` if not specified.
      max_to_keep: Number of checkpoints to keep in the directory. If there are
        more checkpoints than specified by this number, then the oldest
        checkpoints are removed.
      checkpoint_name: Prefix of the checkpoint files (before `-{number}`).
    """
    if max_to_keep < 2:
      raise ValueError("Requires multiple checkpoints (max_to_keep>=2).")
    multihost_base_directory = multihost_base_directory.rstrip("/")
    self.multihost_base_directory = multihost_base_directory
    if host_id is None:
      host_id = jax.host_id()
    base_directory = f"{multihost_base_directory}-{host_id}"
    super().__init__(
        base_directory,
        tf_state,
        max_to_keep=max_to_keep,
        checkpoint_name=checkpoint_name)

  @utils.logged_with(
      "MultihostCheckpoint.get_latest_checkpoint_to_restore_from()")
  def get_latest_checkpoint_to_restore_from(self) -> Optional[str]:
    """Returns the latest checkpoint available on all hosts."""
    base_directory_glob = f"{self.multihost_base_directory}-*"
    base_directories = tf.io.gfile.glob(base_directory_glob)
    if self.base_directory not in base_directories:
      return None
    checkpoints = {}
    common_numbers = None
    for base_directory in base_directories:
      checkpoint_manager = tf.train.CheckpointManager(
          tf.train.Checkpoint(),
          base_directory,
          max_to_keep=self.max_to_keep,
          checkpoint_name=self.checkpoint_name)
      numbers = [
          CheckpointInfo.from_path(checkpoint).number
          for checkpoint in checkpoint_manager.checkpoints
      ]
      checkpoints[base_directory] = dict(
          zip(numbers, checkpoint_manager.checkpoints))
      if common_numbers is None:
        common_numbers = set(numbers)
      else:
        common_numbers &= set(numbers)
    logging.info("Checked checkpoint base_directories: %s - common_numbers=%s",
                 base_directories, common_numbers)
    if not common_numbers:
      return None
    highest_number = sorted(common_numbers)[-1]
    return checkpoints[self.base_directory][highest_number]



# Copyright 2021 The CLU Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for parameter overviews."""

from clu import parameter_overview
from flax import linen as nn
import jax
import jax.numpy as jnp
import sonnet as snt
import tensorflow as tf

EMPTY_PARAMETER_OVERVIEW = """+------+-------+------+------+-----+
| Name | Shape | Size | Mean | Std |
+------+-------+------+------+-----+
+------+-------+------+------+-----+
Total: 0"""

SNT_CONV2D_PARAMETER_OVERVIEW = """+----------+--------------+------+
| Name     | Shape        | Size |
+----------+--------------+------+
| conv/b:0 | (2,)         | 2    |
| conv/w:0 | (3, 3, 3, 2) | 54   |
+----------+--------------+------+
Total: 56"""

SNT_CONV2D_PARAMETER_OVERVIEW_WITH_STATS = """+----------+--------------+------+---------+-------+
| Name     | Shape        | Size | Mean    | Std   |
+----------+--------------+------+---------+-------+
| conv/b:0 | (2,)         | 2    | 0.0     | 0.0   |
| conv/w:0 | (3, 3, 3, 2) | 54   | -0.0127 | 0.157 |
+----------+--------------+------+---------+-------+
Total: 56"""

FLAX_CONV2D_PARAMETER_OVERVIEW = """+-------------+--------------+------+
| Name        | Shape        | Size |
+-------------+--------------+------+
| conv/bias   | (2,)         | 2    |
| conv/kernel | (3, 3, 3, 2) | 54   |
+-------------+--------------+------+
Total: 56"""

FLAX_CONV2D_PARAMETER_OVERVIEW_WITH_STATS = """+-------------+--------------+------+--------+-------+
| Name        | Shape        | Size | Mean   | Std   |
+-------------+--------------+------+--------+-------+
| conv/bias   | (2,)         | 2    | 0.0    | 0.0   |
| conv/kernel | (3, 3, 3, 2) | 54   | 0.0562 | 0.188 |
+-------------+--------------+------+--------+-------+
Total: 56"""

FLAX_CONV2D_MAPPING_PARAMETER_OVERVIEW_WITH_STATS = """+--------------------+--------------+------+--------+-------+
| Name               | Shape        | Size | Mean   | Std   |
+--------------------+--------------+------+--------+-------+
| params/conv/bias   | (2,)         | 2    | 0.0    | 0.0   |
| params/conv/kernel | (3, 3, 3, 2) | 54   | 0.0562 | 0.188 |
+--------------------+--------------+------+--------+-------+
Total: 56"""


class TfParameterOverviewTest(tf.test.TestCase):

  def test_count_parameters_empty(self):
    module = snt.Module()
    snt.allow_empty_variables(module)

    # No variables.
    self.assertEqual(0, parameter_overview.count_parameters(module))

    # Single variable.
    module.var = tf.Variable([0, 1])
    self.assertEqual(2, parameter_overview.count_parameters(module))

  def test_count_parameters_on_module(self):
    module = snt.Module()
    # Weights of a 2D convolution with 2 filters..
    module.conv = snt.Conv2D(output_channels=2, kernel_shape=3, name="conv")
    module.conv(tf.ones((2, 5, 5, 3)))  # 3 * 3*3 * 2 + 2 (bias) = 56 parameters
    self.assertEqual(56, parameter_overview.count_parameters(module))

  def test_count_parameters_on_module_with_duplicate_names(self):
    module = snt.Module()
    # Weights of a 2D convolution with 2 filters..
    module.conv = snt.Conv2D(output_channels=2, kernel_shape=3, name="conv")
    module.conv(tf.ones((2, 5, 5, 3)))  # 3 * 3*3 * 2 + 2 (bias) = 56 parameters
    module.conv2 = snt.Conv2D(output_channels=2, kernel_shape=3, name="conv")
    module.conv2(tf.ones(
        (2, 5, 5, 3)))  # 3 * 3*3 * 2 + 2 (bias) = 56 parameters
    parameter_overview.log_parameter_overview(module)
    self.assertEqual(112, parameter_overview.count_parameters(module))

  def test_get_parameter_overview_empty(self):
    module = snt.Module()
    snt.allow_empty_variables(module)

    # No variables.
    self.assertEqual(EMPTY_PARAMETER_OVERVIEW,
                     parameter_overview.get_parameter_overview(module))

    module.conv = snt.Conv2D(output_channels=2, kernel_shape=3)
    # Variables not yet created (happens in the first forward pass).
    self.assertEqual(EMPTY_PARAMETER_OVERVIEW,
                     parameter_overview.get_parameter_overview(module))

  def test_get_parameter_overview_on_module(self):
    module = snt.Module()
    # Weights of a 2D convolution with 2 filters..
    module.conv = snt.Conv2D(output_channels=2, kernel_shape=3, name="conv")
    module.conv(tf.ones((2, 5, 5, 3)))  # 3 * 3^2 * 2 = 56 parameters
    self.assertEqual(
        SNT_CONV2D_PARAMETER_OVERVIEW,
        parameter_overview.get_parameter_overview(module, include_stats=False))
    self.assertEqual(SNT_CONV2D_PARAMETER_OVERVIEW_WITH_STATS,
                     parameter_overview.get_parameter_overview(module))


class CNN(nn.Module):

  @nn.compact
  def __call__(self, x):
    return nn.Conv(features=2, kernel_size=(3, 3), name="conv")(x)


class JaxParameterOverviewTest(tf.test.TestCase):

  def test_count_parameters_empty(self):
    self.assertEqual(0, parameter_overview.count_parameters({}))

  def test_count_parameters(self):
    rng = jax.random.PRNGKey(42)
    # Weights of a 2D convolution with 2 filters..
    variables = CNN().init(rng, jnp.zeros((2, 5, 5, 3)))
    # 3 * 3*3 * 2 + 2 (bias) = 56 parameters
    self.assertEqual(56,
                     parameter_overview.count_parameters(variables["params"]))

  def test_get_parameter_overview_empty(self):
    self.assertEqual(EMPTY_PARAMETER_OVERVIEW,
                     parameter_overview.get_parameter_overview({}))

  def test_get_parameter_overview(self):
    rng = jax.random.PRNGKey(42)
    # Weights of a 2D convolution with 2 filters..
    variables = CNN().init(rng, jnp.zeros((2, 5, 5, 3)))
    self.assertEqual(
        FLAX_CONV2D_PARAMETER_OVERVIEW,
        parameter_overview.get_parameter_overview(
            variables["params"], include_stats=False))
    print(parameter_overview.get_parameter_overview(variables["params"]))
    self.assertEqual(
        FLAX_CONV2D_PARAMETER_OVERVIEW_WITH_STATS,
        parameter_overview.get_parameter_overview(variables["params"]))
    print(parameter_overview.get_parameter_overview(variables))
    self.assertEqual(
        FLAX_CONV2D_MAPPING_PARAMETER_OVERVIEW_WITH_STATS,
        parameter_overview.get_parameter_overview(variables))


if __name__ == "__main__":
  tf.test.main()
